/* -*- C -*- */

#include "cavity.h"
#include <stdlib.h>
#include <ctype.h>

/*****************************************************
c      subroutine setmin
c
c     Setup Range Minimum/Maximum
c
*****************************************************/

static void setmin(void){
  int i,j;
  for(i=0;i<3;i++){
    cmin[i]=cmax[i]=xyz[iExcl[0]][i];
  }
  for(j=1;j<nExcl;j++){
    for(i=0;i<3;i++){
      double t=xyz[iExcl[j]][i];
      if (t<cmin[i]) cmin[i]=t;
      if (t>cmax[i]) cmax[i]=t;
    }
  }
  for(i=0;i<3;i++){
    cmin[i]-=cAdd;
    cmax[i]+=cAdd;
  }
}

int CavInit(const char *filename) {
  int i,j,k,l;

  if (GlobalInit()==0){
    fprintf(stderr,"Can't allocate memory!\n");
    exit(1);
  };

  l_log=fopen("cav.log","w");
  if (l_log==NULL) error("Can't Open Log File!");
  iErr=0;

  for(i=0;i<MaxAtm;i++){
    iUse[i]=0;
    bExclFlag[i]=0;
  }

  if (!ReadControlFile(filename)) {
    char buf[1024];
    sprintf(buf,"Can't Open Control File '%s'!",filename);
    error(buf);
  }
  if (!ReadDataFile("cav.dat")) error("Can't Open Data File 'cav.dat'!");
  if (!CleanStructure()) error("Memory error!");

  if (!CavOpenPlotter("cav.ps")) error("Can't Open Output File 'cav.ps'!");

  for(i=0;i<nAtom;i++){
    w[i]=Vdw[iTbl[i]-1];
    w2[i]=w[i]*w[i];
  }
  for(i=0;i<nPlt;i++){
    l=iPlt[i];
    for(j=0;j<nExcl;j++){
      k=iExcl[j];
      if (l!=k) {
	double dx,dy,dz,dist;
	dx=xyz[l][0]-xyz[k][0];
	dy=xyz[l][1]-xyz[k][1];
	dz=xyz[l][2]-xyz[k][2];
	dist=dx*dx+dy*dy+dz*dz;
	if (w2[l]>dist) w2[l]=dist;
      }
      w[l]=sqrt(w2[l]);
    }
  }
  for(i=0;i<nAtom;i++){
    fprintf(l_log,"%-6s %5d %12.6f %12.6f %12.6f %10.6f\n",natm[i],i+1,
	    xyz[i][0],xyz[i][1],xyz[i][2],w[i]);
  }
  if (divz<0) divz=0.2;
  fflush(l_log);
  return 1;
}

/****************************************************************
c     initialize2
c     set initial range, axis
****************************************************************/

void cavin2(void){
  int i;
  switch(iView){
  case 1:
    kx=1;
    ky=2;
    kz=0;
    break;
  case 2:
    kx=2;
    ky=0;
    kz=1;
    break;
  case 3:
  default:
    kx=0;
    ky=1;
    kz=2;
    break;
  }
  if (nInt==0) {
    setmin();
  } 
  xmin=cmin[kx];
  xmax=cmax[kx];
  ymin=cmin[ky];
  ymax=cmax[ky];
  zmin=cmin[kz];
  zmax=cmax[kz];
  fprintf(l_log,"# Initial Range\n# %10.3f %10.3f %10.3f %10.3f %10.3f %10.3f %10.3f\n",
	  cmin[0],cmax[0],cmin[1],cmax[1],cmin[2],cmax[2],divz);
  for(i=0;i<3;i++){
    cmin[i]=10000;
    cmax[i]=-10000;
  }
}

/********************************************************
      subroutine used
********************************************************/
void used(void){
  int i;
  for(i=0;i<nAtom;i++){
    if (iUse[i]){
      fprintf(l_log,"%-2s %4d %12.6f %12.6f %12.6f %10.6f\n",natm[i],i,
	      xyz[i][0],xyz[i][1],xyz[i][2],w[i]);
    }
  }
  fprintf(l_log,"# only for volume calc.\n");
  for(i=0;i<nAtom;i++){
    if (iUse[i]==2){
      fprintf(l_log,"%-2s %4d %12.6f %12.6f %12.6f %10.6f\n",natm[i],i,
	      xyz[i][0],xyz[i][1],xyz[i][2],w[i]);
    }
  }
}

/********************************************************
      subroutines for povray file
********************************************************/
static const char *povradii(const char *name){
  if (name[0]=='C' && !isalpha(name[1])) return "radC";
  if (name[0]=='H' && !isalpha(name[1])) return "radH";
  if (name[0]=='N' && !isalpha(name[1])) return "radN";
  if (name[0]=='O' && !isalpha(name[1])) return "radO";
  if (name[0]=='C' && name[1]=='L') return "radCL";
  return "radX";
}

static const char *povcolor(const char *name) {
  if (name[0]=='C' && !isalpha(name[1])) return "catC";
  if (name[0]=='H' && !isalpha(name[1])) return "catH";
  if (name[0]=='N' && !isalpha(name[1])) return "catN";
  if (name[0]=='O' && !isalpha(name[1])) return "catO";
  if (name[0]=='C' && name[1]=='L') return "catCL";
  return "catX";
}

static int povout(void) {
  FILE *pfp;
  double cavr[3],aavr[3];
  int gsc=0,i,j;

  pfp=fopen("cav.pov","w");
  if (pfp==NULL) return 0;

  for(i=0; i<3; i++) {
    aavr[i]=0.0;
  }
  fprintf(pfp,
"#include \"textures.inc\"\n"
"#include \"colors.inc\"\n"
"#include \"metals.inc\"\n"
"\n"
"#declare SCALE = 1.0;\n"
"#declare TRANSLATE = <0,0,0>;\n"
"#declare DISTANCE = 30;\n"
"#declare STEREO = 0;\n"
"#declare STEREOSPLIT = 5;\n"
	  "\n");
  fputs(
"  light_source {< 25, 25, -50> color rgb <1,1,1>  adaptive 1 }\n"
"  light_source {< 0, 0, -100> color rgb <1,1,1>  adaptive 1 shadowless}\n"
"  // light_source {< 0, 0, 100> color rgb <1,1,1>  adaptive 1 shadowless}\n"
"  // light_source {< 0, 100, 0> color rgb <1,1,1>  adaptive 1 shadowless}\n"
"  // light_source {< 0, -100, 0> color rgb <1,1,1>  adaptive 1 shadowless}\n"
,pfp);
  fputs(
"  // light_source {<100, 0, 0> color rgb <1,1,1>  adaptive 1 shadowless}\n"
"  // light_source {<-100, 0, 0> color rgb <1,1,1>  adaptive 1 shadowless}\n"
"  camera { orthographic location <0, 0, -DISTANCE> look_at  <0, 0,  0> }\n"
"  // camera { perspective location <0, 0, -DISTANCE> look_at  <0, 0,  0> }\n"
"  background { color <0.5,0.5,0.5> }\n"
"\n",pfp);

  fprintf(pfp,
"// Element Color\n"
"#declare catC = color rgb <0.5,0.5,0.5>;\n"
"#declare catH = color rgb <0.0,1.0,1.0>;\n"
"#declare catN = color rgb <0.0,0.0,0.7>;\n"
"#declare catO = color rgb <1.0,0.1,0.1>;\n"
"#declare catCL = color rgb <0.0,0.7,0.0>;\n"
"#declare catX = color rgb <0.3,0.3,0.3>;\n"
"// Element Radius\n"
"#declare radC = 0.4;\n"
"#declare radH = 0.2;\n"
"#declare radN = 0.4;\n"
"#declare radO = 0.4;\n"
"#declare radCL = 0.5;\n"
"#declare radX = 0.6;\n"
"// Bond Radius\n"
"#declare bondw = 0.1;\n"
	  );

  for(i=0; i<3; i++) {
    cavr[i]=(cmin[i]+cmax[i])*0.5;
  }
  /*
  fprintf(pfp,
	  "#declare RANGEBOX =\n"
	  "  intersection {\n"
	  "    plane { -x, %.3f }\n"
	  "    plane { x, %.3f }\n"
	  "    plane { -y, %.3f }\n"
	  "    plane { y, %.3f }\n"
	  "    plane { -z, %.3f }\n"
	  "    plane { z, %.3f }\n"
	  "  }\n",
	  -cmin[0], cmax[0], -cmin[1], cmax[1], -cmin[2], cmax[2]);
	  */
  fprintf(pfp,
	  "#declare RANGEBOX = box { <%.3f,%.3f,%.3f>,<%.3f,%.3f,%.3f> } \n",
	  cmin[0], cmin[1], cmin[2], cmax[0], cmax[1], cmax[2]);
  fprintf(pfp,
	  "#declare RANGEBOX_Center =<%f,%f,%f>;\n",cavr[0],cavr[1],cavr[2]);

  fprintf(pfp, /* Dummy Range Box */
  "// object { RANGEBOX pigment {color rgbt <0.8,0.6,0.8,0.9>}} // Range\n");

  if (sradi>0) {
    fprintf(pfp,
	    "#declare SPHERE = sphere { <%.3f,%.3f,%.3f>, %.3f }\n",
	    sxyz[0],sxyz[1],sxyz[2],sradi);
  }
  fprintf(pfp,
	  "#declare CAVITY =\n"
	  " difference {\n"
	  "  object { RANGEBOX }\n"
	  "  merge {\n");
  for(i=0;i<nAtom;i++){
    if (iUse[i]){
      fprintf(pfp, "   sphere { <%.3f,%.3f,%.3f>, %.3f } // %s %d\n",
	      xyz[i][0],xyz[i][1],xyz[i][2],w[i],natm[i],i);
    }
  }

  fprintf(pfp,
	  "  }\n"
	  "  pigment {color rgbt <0.8,1.0,1.0,0.5>} // Color for Cavity\n"
	  " }\n"
	  "// end of CAVITY\n");


  fprintf(pfp, "#declare ATOMS =\n"
	  " union {\n");

  for(i=0;i<nPlt;i++){
    int k;
    k=iPlt[i];
    fprintf(pfp, "   sphere { <%.3f,%.3f,%.3f>, %s pigment {%s}}\n",
	    xyz[k][0],xyz[k][1],xyz[k][2],
	    povradii(natm[k]),povcolor(natm[k]));
    for(j=0;j<3;j++) {
      aavr[j]+=xyz[k][j];
    }
    gsc++;
  }

  fprintf(pfp, " } // end of Atoms\n");
  fprintf(pfp, "#declare MODEL =\n"
	  " union {\n");

  for(i=0;i<nBond;i++){
    double *sxyz,*exyz;
    sxyz=xyz[iPs[i]];
    exyz=xyz[iPe[i]];
    for(j=0;j<3;j++) cavr[j]=(sxyz[j]+exyz[j])*0.5;
    fprintf(pfp, 
    "  cylinder { <%.3f,%.3f,%.3f>,<%.3f,%.3f,%.3f>,bondw pigment {%s} }\n",
	    cavr[0],cavr[1],cavr[2],
	    sxyz[0],sxyz[1],sxyz[2],povcolor(natm[iPs[i]]));
    fprintf(pfp, 
    "  cylinder { <%.3f,%.3f,%.3f>,<%.3f,%.3f,%.3f>,bondw pigment {%s} }\n",
	    cavr[0],cavr[1],cavr[2],
	    exyz[0],exyz[1],exyz[2],povcolor(natm[iPe[i]]));
  }

  fprintf(pfp, " } // end of Model\n");

  fprintf(pfp,"#declare CENTER=<%.3f,%.3f,%.3f>;\n",
	  aavr[0]/gsc,aavr[1]/gsc,aavr[2]/gsc);

  fprintf(pfp, 
	  "#declare FULL=\n"
	  " union {\n"
	  "  object {MODEL} \n"
	  "  intersection { \n"
	  "   object {CAVITY} \n"
	  );
  if (sradi>0) {
    fprintf(pfp, 
	    "   object {SPHERE} \n");
  }
  fprintf(pfp, 
          "  }\n"
	  "//  object {ATOMS} \n"
	  "//  object {RANGEBOX pigment { rgbt<0.7,0.3,0.3,0.8>}}\n"
	  "//  object {SPHERE pigment { rgbt<0.7,0.7,0.3,0.8>}}\n"
	  "  translate -CENTER\n"
	  "  rotate <0,clock*360,0>\n"
	  "  scale SCALE\n"
	  "  translate TRANSLATE\n"
	  " }\n");
  fprintf(pfp,
	  "#if (STEREO)\n"
	  "  object {FULL rotate <0,5,0> translate <-STEREOSPLIT,0,0> }\n"
	  "  object {FULL rotate <0,-5,0> translate <STEREOSPLIT,0,0> }\n"
	  "#else\n"
	  "  object {FULL}\n"
	  "#end\n"
	  );


  fclose(pfp);
  pfp=fopen("cav.ini","r");
  if (pfp!=NULL) {
    fclose(pfp);
  } else {
    pfp=fopen("cav.ini","w");
    if (pfp!=NULL) {
      fprintf(pfp,
	      "Width=320\n"
	      "Height=240\n"
	      "Antialias=Off\n"
	      "Antialias_Threshold=0.3\n"
	      "Antialias_Depth=3\n"
	      "Input_File_Name=cav.pov\n"
	      "Initial_Frame=1\n"
	      "Final_Frame=50\n"
	      "Initial_Clock=0\n"
	      "Final_Clock=1\n"
	      "Pause_when_Done=off\n"
	      "Cyclic_Animation=off\n");
      fclose(pfp);
    }
  }
  return 1;
}

/*****************************************************************
c
c     Cavity Final Message Print Routine
c     (Range, Step, Volume Print Out)
c     And export povray file.
c
*****************************************************************/

int CavTrailer(void){
  int i;
  for(i=0;i<3;i++){
    cmin[i]-=2*divz;
    cmax[i]+=2*divz;
  }
  if (cmin[kx]<xmin || nInt!=0) cmin[kx]=xmin;
  if (cmax[kx]>xmax || nInt!=0) cmax[kx]=xmax;
  if (cmin[ky]<ymin || nInt!=0) cmin[ky]=ymin;
  if (cmax[ky]>ymax || nInt!=0) cmax[ky]=ymax;
  if (cmin[kz]<zmin || nInt!=0) cmin[kz]=zmin;
  if (cmax[kz]>zmax || nInt!=0) cmax[kz]=zmax;
  fprintf(l_log,"$ Range\n");
  fprintf(l_log,"  XMIN,  XMAX,  YMIN,  YMAX,"
	  "  ZMIN,  ZMAX,   DIV\n");
  fprintf(l_log,"%7.3f%7.3f%7.3f%7.3f%7.3f%7.3f%7.3f\n",
	  cmin[0],cmax[0],cmin[1],cmax[1],cmin[2],cmax[2],divz);
  sprintf(lbuf,"XRANGE = %10.4f%10.4f",cmin[0],cmax[0]);
  CavPlotComment(lbuf);
  sprintf(lbuf,"YRANGE = %10.4f%10.4f",cmin[1],cmax[1]);
  CavPlotComment(lbuf);
  sprintf(lbuf,"ZRANGE = %10.4f%10.4f",cmin[2],cmax[2]);
  CavPlotComment(lbuf);
  if (iErr==0){
    if (iVol!=0){
      fprintf(l_log,"Volume =  %10.3f\n",vol);
      sprintf(lbuf,"Volume = %8.3f",vol);
      CavPlotComment(lbuf);
    } else {
      fprintf(l_log,"Volume =  %10.3f\n",-1.0);
    }
  } else {
    sprintf(lbuf,"Cavity Calculation Error\n");
    CavPlotComment(lbuf);
  }
  fflush(l_log);
  fprintf(l_log,"$ CavAtom\n");
  used();
  fflush(l_log);
  fprintf(l_log,"$ PlotAtom\n");
  for(i=0;i<nPlt;i++){
    int k;
    k=iPlt[i];
    fprintf(l_log,"%-6s %5d %12.5f %12.5f %12.5f %12.5f\n",natm[k],iTbl[k],
	    xyz[k][0],xyz[k][1],xyz[k][2],w[k]);
  }
  fflush(l_log);
  fprintf(l_log,"$ CavBond\n");
  for(i=0;i<nBond;i++){
    double *sxyz,*exyz;
    sxyz=xyz[iPs[i]];
    exyz=xyz[iPe[i]];
    fprintf(l_log,"%-6s %12.5f %12.5f %12.5f %-6s %12.5f %12.5f %12.5f \n",
	    natm[iPs[i]],sxyz[0],sxyz[1],sxyz[2],
	    natm[iPe[i]],exyz[0],exyz[1],exyz[2]);
  }
  fprintf(l_log,"$ End\n");
  fflush(l_log);
  CavClosePlotter();

  if (iErr==0) povout();

  fclose(l_log);
  return 1;
}
